import os
import warnings

warnings.filterwarnings("ignore", category=DeprecationWarning)
warnings.filterwarnings('ignore', message='Unverified HTTPS request')
warnings.filterwarnings('ignore', message='RequestsDependencyWarning')

lazyops_root = os.path.abspath(os.path.dirname(__file__))

from .envs import LazyEnv, get_logger
from .models import LazyData, LazyTime, LazyDate, LazyFormatter, LazyTimer
from .common import lazylibs, lazy_init, run_cmd, clone_repo, File
from .utils import find_binary_in_path, timed_cache, latest_tf_ckpt, require_module
from .utils import build_chunks, retryable, list_to_dict
from .apis import LazyAPI, LazyAPIConfig

env = LazyEnv
item = LazyData
timer = LazyTimer
ttime = LazyTime
formatter = LazyFormatter
dtime = LazyDate
libs = lazylibs
fio = File
lazyapi = LazyAPI
apiconfig = LazyAPIConfig



import sys
import threading
import logging
from dataclasses import dataclass
from typing import Optional, Any, Dict

try:
    from google.colab import auth
    _colab = True
except ImportError:
    _colab = False

_notebook = sys.argv[-1].endswith('json')

_logging_levels = {
    'debug': logging.DEBUG,
    'info': logging.INFO,
    'warn': logging.WARN,
    'warning': logging.WARNING,
    'error': logging.ERROR,
}

class ThreadSafeHandler:
    def __init__(
        self, 
        lock_mode: str = 'lock',
        handler: Optional[Any] = None):
        self.lock_mode = lock_mode
        self.handler = handler
        self.lock = threading.RLock()  if self.lock_mode == 'rlock' else threading.Lock()

    def get(self, init_func: function = None, *args, **kwargs):
        if self.lock_mode == 'rlock': self.lock.acquire()
        with self.lock:
            if not self.handler:
                self.handler = init_func(*args, **kwargs)
            return self.handler


class LogFormatter(logging.Formatter):
    COLOR_CODES = {
        logging.CRITICAL: "\033[38;5;196m", # bright/bold magenta
        logging.ERROR:    "\033[38;5;9m", # bright/bold red
        logging.WARNING:  "\033[38;5;11m", # bright/bold yellow
        logging.INFO:     "\033[38;5;111m", # white / light gray
        logging.DEBUG:    "\033[1;30m"  # bright/bold black / dark gray
    }

    RESET_CODE = "\033[0m"
    def __init__(self, color, *args, **kwargs):
        super(LogFormatter, self).__init__(*args, **kwargs)
        self.color = color

    def format(self, record, *args, **kwargs):
        if (self.color == True and record.levelno in self.COLOR_CODES):
            record.color_on  = self.COLOR_CODES[record.levelno]
            record.color_off = self.RESET_CODE
        else:
            record.color_on  = ""
            record.color_off = ""
        return super(LogFormatter, self).format(record, *args, **kwargs)

class LazyOpsLogger:
    def __init__(self, config):
        self.config = config
        self.logger = self.setup_logging()
    
    def setup_logging(self):
        logger = logging.getLogger(self.config['name'])
        logger.setLevel(_logging_levels[self.config.get('log_level', 'info')])

        console_log_output = sys.stdout if _notebook or _colab else sys.stderr        
        console_handler = logging.StreamHandler(console_log_output)
        console_handler.setLevel(self.config["console_log_level"].upper())
        console_formatter = LogFormatter(fmt=self.config["log_line_template"], color=self.config["console_log_color"])
        console_handler.setFormatter(console_formatter)
        if self.config.get('clear_handlers', False) and logger.hasHandlers():
            logger.handlers.clear()
        logger.addHandler(console_handler)
        if self.config.get('quiet_loggers'):
            to_quiet = self.config['quiet_loggers']
            if isinstance(to_quiet, str): to_quiet = [to_quiet]
            for clr in to_quiet:
                clr_logger = logging.getLogger(clr)
                clr_logger.setLevel(logging.ERROR)

        logger.propagate = self.config.get('propagate', False)
        return logger

    def get_logger(self):
        return self.logger
    
    def debug(self, msg, *args, **kwargs):
        return self.logger.debug(msg, *args, **kwargs)
    
    def info(self, msg, *args, **kwargs):
        return self.logger.info(msg, *args, **kwargs)
    
    def warn(self, msg, *args, **kwargs):
        return self.logger.warn(msg, *args, **kwargs)
    
    def error(self, msg, *args, **kwargs):
        return self.logger.error(msg, *args, **kwargs)
    
    def __call__(self, msg, *args, **kwargs):
        return self.logger.info(msg, *args, **kwargs)


def setup_new_logger(name, log_level='info', quiet_loggers=None, clear_handlers=False, propagate=True):
    logger_config = {
        'name': name,
        'log_level': log_level,
        'console_log_output': "stdout", 
        'console_log_level': "info",
        'console_log_color': True,
        'logfile_file': None,
        'logfile_log_level': "debug",
        'logfile_log_color': False,
        'log_line_template': f"%(color_on)s[{name}] %(funcName)-5s%(color_off)s: %(message)s",
        'clear_handlers': clear_handlers,
        'quiet_loggers': quiet_loggers,
        'propagate': propagate
    }
    return LazyOpsLogger(logger_config)


class EnvChecker:
    is_colab = _colab
    is_notebook = _notebook
    is_jpy = _notebook
    loggers: Optional[Dict[str, ThreadSafeHandler]] = {}

    def get_logger(self, name = 'LazyOps', *args, **kwargs):
        if not EnvChecker.loggers.get(name):
            EnvChecker.loggers[name] = ThreadSafeHandler()
        return EnvChecker.loggers[name].get(setup_new_logger, *args, **kwargs)
    

LazyEnv = EnvChecker()
get_logger = LazyEnv.get_logger
logger = get_logger()




import requests
import aiohttp
import asyncio

from dataclasses import dataclass
from dataclasses_json import dataclass_json
from typing import List, Dict, Any, Optional

async def async_req(sess, url, decode_json=True, with_url=False, *args, **kwargs):
    async with sess.request(url, *args, **kwargs) as resp:
        data = await resp.json() if decode_json else await resp
        if not with_url:
            return data
        return {'url': url, 'data': data}


class LazySession:
    def __init__(self, header=None):
        self.header = header
        self.sess = requests.session()
        if header:
            self.sess.headers.update(header)
    
    def fetch(self, decode_json=True, *args, **kwargs):
        res = self.sess.request(*args, **kwargs)
        if decode_json:
            return res.json()
        return res
    
    async def async_batch(self, url, batch_params, decode_json=True, *args, **kwargs):
        tasks = []
        async with aiohttp.ClientSession(headers=self.header) as sess:
            for batch in batch_params:
                tasks.append(asyncio.ensure_future(async_req(sess=sess, url=url, *args, **batch, **kwargs)))
            all_tasks = await asyncio.gather(*tasks)
            return [task for task in all_tasks]
    
    async def async_fetch_urls(self, urls, decode_json=True, *args, **kwargs):
        tasks = []
        async with aiohttp.ClientSession(headers=self.header) as sess:
            for url in urls:
                tasks.append(asyncio.ensure_future(async_req(sess=sess, url=url, with_url=True, *args, **kwargs)))
            all_tasks = await asyncio.gather(*tasks)
            return [task for task in all_tasks]

    async def async_fetch(self, decode_json=True, *args, **kwargs):
        async with aiohttp.ClientSession(headers=self.header) as sess:
            async with sess.request(*args, **kwargs) as resp:
                return await resp.json() if decode_json else await resp

    def __call__(self, *args, **kwargs):
        return self.fetch(*args, **kwargs)
    
    def __exit__(self, _):
        self.sess.close()



@dataclass_json
@dataclass
class LazyRoute:
    path: str
    method: Optional[str] = 'POST'
    data_key: Optional[str] = 'inputs'
    params_key: Optional[str] = 'params'
    params: Optional[Dict[str, Any]] = None
    prefix_payload: Optional[bool] = True
    is_async: Optional[bool] = False
    decode_json: Optional[bool] = True

    def get_config(self, base_url, data=None, **config):
        p = self.params.copy() if self.params else {}
        if config:
            for k,v in config.items():
                if ((self.params and k in self.params) or not self.params) and v:
                    p[k] = v
        if base_url.endswith('/'):
            base_url = base_url[:-1]
        if not self.path.startswith('/'):
            self.path = '/' + self.path
        pc = {self.params_key: p} if self.params_key else p
        if data: pc[self.data_key] = data
        if self.prefix_payload: pc = {'payload': pc}
        return {'method': self.method, 'url': base_url + self.path, self.pkey: pc, 'decode_json': self.decode_json}

    @property
    def pkey(self):
        if self.method in ['POST']:
            return 'json'
        if self.method in ['GET']:
            return 'params'
        return 'data'



@dataclass_json
@dataclass
class LazyHeader:
    user: Optional[str] = None
    key: Optional[str] = None
    token: Optional[str] = None
    content_type: Optional[str] = 'application/json'

    @property
    def config(self):
        h = {'Content-type': self.content_type}
        if self.token:
            h['Authorization'] = f'Bearer {self.token}'
        elif self.user and self.key:
            h.update({'username': self.user, 'password': self.key})
        return h


@dataclass_json
@dataclass
class LazyAPIModel:
    url: str
    header: LazyHeader
    routes: Dict[str, LazyRoute]
    session: Optional[LazySession] = None

    @property
    def sess(self):
        if not self.session:
            self.session = LazySession(header=self.header.config)
        return self.session

    async def async_call(self, route, *args, **kwargs):
        return await self.sess.async_fetch(**self.routes[route].get_config(base_url=self.url, *args, **kwargs))

    def get(self, route, *args, **kwargs):
        if route not in self.routes:
            raise ValueError
        if self.routes[route].is_async and kwargs.get('call_async', False):
            return self.async_call(route, *args, **kwargs)
        return self.sess(**self.routes[route].get_config(base_url=self.url, *args, **kwargs))

    def __call__(self, route, *args, **kwargs):
        return self.get(route, *args, **kwargs)

@dataclass_json
@dataclass
class LazyAPIConfig:
    url: str
    user: Optional[str] = None
    key: Optional[str] = None
    token: Optional[str] = None
    default_params: Optional[Dict[str, str]] = None
    params_key: Optional[str] = None
    data_key: Optional[str] = None
    default_fetch: Optional[str] = None
    default_async: Optional[str] = None
    route_config: Optional[Dict[str, LazyRoute]] = None


class LazyAPI:
    def __init__(self, config: LazyAPIConfig, *args, **kwargs):
        self.config = config
        self.header = LazyHeader(user=self.config.user, key=self.config.key, token=self.config.token)
        self.api = LazyAPIModel(self.config.url, self.header, self.config.route_config)
        for route_name in list(self.config.route_config.keys()):
            func = (lambda route_name=route_name, *args, **kwargs: self.api.get(route=route_name, *args, **kwargs))
            setattr(self, route_name, func)
    
    def api_call(self, is_async=False, *args, **kwargs):
        return self.api(route=(self.config.default_async if is_async else self.config.default_fetch), *args, **kwargs)
    
    async def async_call(self, *args, **kwargs):
        return await self.api.async_call(route=self.config.default_async, *args, **kwargs)
    
    def __call__(self, route, *args, **kwargs):
        return self.api(route, *args, **kwargs)
    
    @classmethod
    def build(cls, config, *args, **kwargs):
        if isinstance(config, LazyAPIConfig):
            pass
        elif isinstance(config, dict):
            config = LazyAPIConfig.from_dict(config)
        elif isinstance(config, str):
            config = LazyAPIConfig.from_json(config)
        return LazyAPI(config, *args, **kwargs)






import os 
import sys
from setuptools import setup, find_packages

version = '0.0.1'
binary_names = ['lazyops']
pkg_name = 'lazyops'

root = os.path.abspath(os.path.dirname(__file__))
packages = find_packages(include=[pkg_name, "{}.*".format(pkg_name)])


deps = {
    'main': [
        'file-io>=0.1.15',
        'requests',
        'dateparser',
        'dataclasses-json',
        'aiohttp',
        'six'
    ]
}

with open(os.path.join(root, 'README.md'), 'rb') as readme:
    long_description = readme.read().decode('utf-8')

setup(
    name=pkg_name,
    version=version,
    description=pkg_name,
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Tri Songz',
    author_email='ts@growthengineai.com',
    url=f'http://github.com/trisongz/{pkg_name}',
    python_requires='>=3.7',
    install_requires=deps['main'],
    packages=packages,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Intended Audience :: Developers",
    ]
)